import logging


def chunks(content, maxlength):
    """Yield successive n-sized chunks from lst."""
    start = 0
    end = 0
    if not content or len(content) == 0:
        return ""
    while start + maxlength < len(content) and end != -1:
        end = content.rfind(" ", start, start + maxlength + 1)
        yield content[start:end]
        start = end + 1
    yield content[start:]


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(
        logging.Formatter(
            "[%(asctime)s.%(msecs)03d] [%(process)s] (%(name)s) [%(levelname)s] - %(message)s"
        )
    )
    logger.addHandler(stream_handler)
    return logger



import os
import time
from pathlib import Path
from typing import Optional

from fastapi import FastAPI
from fastapi.responses import JSONResponse
from qdrant_client.http import models

from simgen_ssg import utils
from simgen_ssg.parsers import parser_for_file

logger = utils.get_logger(__name__)

app = FastAPI()
app.ready = False
app.qdrant_server = None


@app.get("/")
async def index():
    response = {"ready": app.ready}
    if not app.ready:
        return JSONResponse(response)
    q_client = app.qdrant_server.qdrant_client
    collection_instance = await q_client.get_collection(app.qdrant_server.QDRANT_COLLECTION_NAME)
    response.update({"vectors": collection_instance.vectors_count})
    return JSONResponse(response)


@app.post("/add")
async def save_content(id: str, body: str, collection_name: str):
    await app.qdrant_server.save_content_to_db(id, body, collection_name, time.time())


@app.get("/recommend")
async def get_recommendations(
    collection: Optional[str],
    id: Optional[str] = "",
    q: Optional[str] = "",
    limit: Optional[int] = 5,
):
    """
    Add vectors to the collection
    """
    q_client = app.qdrant_server.qdrant_client
    content = ""
    if id:
        file_path = id
        logger.info(f"Getting recommendations for {file_path}")
        for dir in app.content_dirs:
            new_path = os.path.join(dir, id)
            if os.path.exists(new_path):
                file_path = os.path.abspath(new_path)
                break
        else:
            return JSONResponse({"error": "File1 not found"}, status_code=404)
        try:
            parser = parser_for_file(Path(file_path))
            content = parser.content
        except FileNotFoundError:
            # Return 404
            return JSONResponse({"error": "File2 not found"}, status_code=404)
    elif q:
        content = q
    qdrant_result = await q_client.query(
        collection_name=app.qdrant_server.QDRANT_COLLECTION_NAME,
        query_text=content,
        query_filter=models.Filter(
            must_not=[
                models.FieldCondition(
                    key="file_path",
                    match=models.MatchValue(value=id),
                ),
            ],
            must=[
                models.FieldCondition(
                    key="collection",
                    match=models.MatchValue(value=collection),
                )
            ],
        ),
        limit=limit,
    )
    response = list()
    logger.debug(f"Found {len(qdrant_result)} results")
    for result in qdrant_result:
        result.metadata.update({"score": result.score})
        result.metadata.pop("document")
        response.append(result.metadata)
    # Convert response list of dictionary to unique by file_path
    unique_response = []
    unique_keys = set()
    for item in response:
        key = item.get("file_path")
        if key not in unique_keys:
            unique_response.append(item)
            unique_keys.add(key)
    return JSONResponse(unique_response)



import asyncio
import os
import sqlite3
import sys
from pathlib import Path
from typing import List, Optional

from hypercorn.asyncio import serve as hypercorn_serve
from hypercorn.config import Config
from qdrant_client import AsyncQdrantClient

from simgen_ssg import simgen_http as simgen_http
from simgen_ssg import utils
from simgen_ssg.parsers import parser_for_file
from simgen_ssg.watcher import files_updated_since

logger = utils.get_logger(__name__)


class QdrantServer:
    QDRANT_COLLECTION_NAME = "simgen"

    def __init__(self, cache_dir, model):
        self.cache_dir = cache_dir
        self.model = model
        self.qdrant_client: Optional[AsyncQdrantClient] = None
        self.sql_client: Optional[sqlite3.Connection] = None

    async def setup_qdrant(self, force=False):
        if force or self.qdrant_client is None:
            cache_dir = Path(self.cache_dir)
            cache_dir.mkdir(parents=True, exist_ok=True)
            qdrant_db_path = cache_dir.joinpath("qdrant_db")
            sqlite_db_path = cache_dir.joinpath("indexes.db")
            os.makedirs(qdrant_db_path, exist_ok=True)
            self.qdrant_client = AsyncQdrantClient(path=qdrant_db_path)
            self.sql_client = sqlite3.connect(sqlite_db_path)

            self.sql_client.execute(
                """
            CREATE TABLE IF NOT EXISTS indexes (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                file_path TEXT NOT NULL,
                chunk_id INTEGER NOT NULL,
                updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                UNIQUE(file_path, chunk_id)
            );
            """
            )
            return True
        return False

    async def fetch_model(self):
        await self.setup_qdrant()
        if self.qdrant_client is None:
            raise Exception("Qdrant client not initialized")
        self.qdrant_client.set_model(self.model)
        existing_collections = await self.qdrant_client.get_collections()
        logger.info(f"Existing collections: {existing_collections.collections}")
        if all(
            [
                collection.name != self.QDRANT_COLLECTION_NAME
                for collection in existing_collections.collections
            ]
        ):
            logger.info(
                f"Could not find collection: {self.QDRANT_COLLECTION_NAME}. Creating new collection"
            )
            # TODO: Check the request model and create collection if the vector_config is different
            await self.qdrant_client.create_collection(
                collection_name=self.QDRANT_COLLECTION_NAME,
                vectors_config=self.qdrant_client.get_fastembed_vector_params(),
            )

    async def save_content_to_db(self, id, content, collection_name, m_time):
        content_chunks = list(utils.chunks(content, 500))
        # Batch insert into both sqlite and qdrant
        logger.debug(f"Saving content to db: {id}. Found chunks: {len(content_chunks)}")
        if self.sql_client is None or self.qdrant_client is None:
            raise Exception("Databases not initialized")
        self.sql_client.execute(
            "INSERT OR IGNORE INTO indexes (file_path, chunk_id, updated_at) VALUES {};".format(
                ",".join(
                    [f"('{id}', {index}, {m_time})" for index in range(1, len(content_chunks) + 1)]
                )
            )
        )
        self.sql_client.commit()
        res = self.sql_client.execute(
            f"SELECT * FROM indexes where file_path='{id}' ORDER BY chunk_id"
        ).fetchall()
        if len(res) > len(content_chunks):
            # The existing content sync'd is more than the new content. Delete the extra content
            logger.debug(f"Deleting extra content for ID: {id}")
            self.sql_client.execute(
                f"DELETE FROM indexes where file_path='{id}' AND chunk_id>{len(content_chunks)}"
            )
            res = res[: len(content_chunks)]
        logger.debug(f"Saving content to qdrant: {id}")
        await self.qdrant_client.add(
            collection_name=self.QDRANT_COLLECTION_NAME,
            documents=content_chunks,
            ids=[instance[0] for instance in res],
            metadata=[
                {"file_path": id, "collection": collection_name, "id": instance[0]}
                for instance in res
            ],
        )

    async def _file_watcher(self, dir_path):
        if self.sql_client is None:
            raise Exception("Databases not initialized")
        res = self.sql_client.execute(
            "SELECT updated_at from indexes ORDER BY updated_at DESC LIMIT 1"
        ).fetchone()
        last_mtime = res[0] if (res and len(res) > 0) else 0
        logger.info(f"Last updated time: {last_mtime}")
        for file_path, rel_path, file_mtime in files_updated_since(dir_path, last_mtime):
            file_path = Path(file_path)
            logger.debug(f"File updated {rel_path}. Last updated time: {file_mtime}")

            rel_path = os.path.relpath(file_path, dir_path)
            parser = parser_for_file(file_path)
            logger.debug(f"Found parser for the file: {rel_path}. {parser.__class__}")
            await self.save_content_to_db(
                rel_path, parser.content, file_path.parent.name, file_mtime
            )

    async def serve(self, directories: List[str], queue: asyncio.Queue, watch: bool = False):
        await queue.put({"ready": False})
        logger.info("Initializing File watcher. This may take a while...")
        await asyncio.gather(*[self._file_watcher(dir) for dir in directories])
        await queue.put({"ready": True})


async def mark_ready(app, queue):
    while True:
        resp = await queue.get()
        if isinstance(resp, dict) and "ready" in resp:
            app.ready = resp["ready"]
            if app.ready:
                return


async def run(directories, bind, watch, cache_dir, model):
    app = simgen_http.app
    qdrant_server = QdrantServer(cache_dir=cache_dir, model=model)
    await qdrant_server.setup_qdrant()
    await qdrant_server.fetch_model()
    app.content_dirs = directories
    app.qdrant_server = qdrant_server
    try:
        hypercorn_config = Config()
        hypercorn_config.bind = bind
        queue: asyncio.Queue = asyncio.Queue()
        app.queue = queue
        await asyncio.gather(
            hypercorn_serve(app, hypercorn_config),
            mark_ready(app, queue),
            qdrant_server.serve(directories, queue, watch),
        )  # fastAPI  # Qdrant file watcher
    except KeyboardInterrupt:
        sys.exit(1)



import logging
import os
import re
from pathlib import Path

blacklist = ["^\.", "\.swp$"]
whitelist = ["[^_].*\.md$"]

fw_logger = logging.getLogger("file_watch")
fw_logger.setLevel(logging.DEBUG)


def file_filter(name):
    def run_filters(name, filters):
        for regex in filters:
            if re.search(regex, name):
                return True
        return False

    if len(whitelist) > 0:
        return run_filters(name, whitelist)
    else:
        return not run_filters(name, blacklist)


def files_updated_since(path, last_mtime):
    req_path = Path(path)
    if not req_path.exists():
        raise FileNotFoundError("Path %s does not exist" % path)
    if not req_path.is_dir():
        raise NotADirectoryError("Path %s is not a directory" % path)
    # assert False, os.walk(req_path)
    for root, _, files in os.walk(req_path):
        for file_name in filter(file_filter, files):
            file_path = Path(root).joinpath(file_name)
            if not os.path.isfile(file_path):
                continue
            file_mtime = os.path.getmtime(file_path)
            if file_mtime > last_mtime:
                yield (file_path, os.path.relpath(file_path, req_path), file_mtime)



from abc import ABC, abstractmethod

# , abstractproperty


class AbstractBaseParser(ABC):
    def __init__(self, file_path):
        self.file_path = file_path
        self.last_processed = None
        self._content = None

    @abstractmethod
    def parse(self):
        pass

    @property
    def content(self):
        if self._content is None:
            self.parse()
        return self._content


class BaseParser(AbstractBaseParser):
    def parse(self):
        pass



from simgen_ssg.parsers.base import AbstractBaseParser


class MarkdownParser(AbstractBaseParser):
    def parse(self):
        with open(self.file_path, "r") as f:
            self._content = f.read()



import logging

from simgen_ssg.cli import cli

fw_logger = logging.getLogger("file_watch")
http_logger = logging.getLogger("http")
fw_logger.setLevel(logging.INFO)
http_logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(
    logging.Formatter(
        "[%(asctime)s.%(msecs)03d] [%(process)s] (%(name)s) [%(levelname)s] - %(message)s"
    )
)
fw_logger.addHandler(stream_handler)


# async def file_watcher(should_watch=False, dirs=[]):
#     fw_logger.info("Initializing File watcher. This may take a while...")
#     await asyncio.gather(*[_file_watcher(dir) for dir in dirs])
#     app.ready = True
#     fw_logger.info("Files synchronized. The server is ready to accept connections")
#     if should_watch:
#         fw_logger.info("Watching files for changes")
#         while True:
#             await asyncio.sleep(3)
#             await asyncio.gather(*[_file_watcher(dir) for dir in dirs])


if __name__ == "__main__":
    cli(_anyio_backend="asyncio")



import asyncclick as click

from simgen_ssg import server


@click.group()
async def cli():
    pass


@cli.command()
@click.option("--cache-dir", "-c", default="./.simgen_cache", help="Cache directory")
@click.option("--model", "-m", default="BAAI/bge-small-en", help="Model name")
async def fetch_model(cache_dir, model):
    """
    Set the model to use for embeddings
    """
    qdrant_server = server.QdrantServer(cache_dir=cache_dir, model=model)
    await qdrant_server.setup_qdrant()
    await qdrant_server.fetch_model()


@cli.command()
@click.option("--dir", "-d", default=["."], help="Content directory", multiple=True)
@click.option("--bind", "-b", default="127.0.0.1:8000", help="Content directory")
@click.option("--watch", "-w", default=False, help="Content directory", is_flag=True)
@click.option("--cache-dir", "-c", default="./.simgen_cache", help="Cache directory")
@click.option("--model", "-m", default="BAAI/bge-small-en", help="Model name")
async def serve(dir, bind, watch, cache_dir, model):
    """
    SimGen SSG: A reccomendation engine for static site generators
    """

    await server.run(directories=dir, bind=bind, watch=watch, cache_dir=cache_dir, model=model)



# import asyncio
# from unittest.mock import MagicMock

# from simgen_ssg.__main__ import main


# async def test_main():
#     # Mock the setup_qdrant function
#     main.setup_qdrant = MagicMock()

#     # Mock the serve and file_watcher coroutines
#     main.serve = MagicMock()
#     main.file_watcher = MagicMock()

#     # Call the main function
#     await main.main()

#     # Assert that the setup_qdrant function was called
#     main.setup_qdrant.assert_called_once()

#     # Assert that the serve and file_watcher coroutines were called with the correct arguments
#     main.serve.assert_called_once_with(main.app, main.Config())
#     main.file_watcher.assert_called_once()

#     # Assert that the serve and file_watcher coroutines were called concurrently
#     asyncio.gather.assert_called_once_with(main.serve(), main.file_watcher())



from pathlib import Path

from simgen_ssg.parsers.base import BaseParser
from simgen_ssg.parsers.markdown import MarkdownParser


def parser_for_file(file_path: Path) -> BaseParser:
    def identify_parser(file_path: Path):
        if file_path.is_file():
            _, ext = file_path.name.rsplit(".", 1)
            if ext == "md":
                return MarkdownParser
            else:
                return BaseParser
        raise Exception("Not a file")

    return identify_parser(file_path)(file_path)
